
-------------------------------------------------------------------------
-- Load :: チャンク(luacの出力)文字列からコードをロード
--
-- オブジェクトの表現
--   nil,bool,number,string,table,組み込み関数 はホスト関数の値を使用
--   普通の関数は特別なテーブルで表現。こんな感じ
--    { type   = type_function (通常のtableと区別するためのフラグ)
--      numup  = upvalueの数
--      numpr  = 引数の数
--      varflg = 可変長引数フラグ
--      maxstk = 最大スタック使用量
--      code   = 命令の配列
--      kv     = 定数値のテーブル
--      kp     = 関数プロトタイプのテーブル
--      upval  = UpValueの配列
--      env    = 環境 }
--   コルーチンはこんな。
--    { type       = type_thread (通常のtableと区別するためのフラグ)
--      dstk, dtop = データスタック
--      cstk, ctop = コールスタック }
-------------------------------------------------------------------------

type_function  = {}
type_coroutine = {}

function bits(n, lsb, numbits) -- ビット演算ユーティリティ
  return math.floor(n / 2^lsb) % 2^numbits
end

function inst_decode(i) -- 命令デコーダ
  return { rawval = i,
           OP = bits(i, 0, 6),
            A = bits(i, 6, 8),
            C = bits(i,14, 9),
            B = bits(i,23, 9),
            Bx= bits(i,14,18),
           sBx= bits(i,14,18) - 131071 }
end

function load_chunk(str)
  if str:sub(1,4) ~= "\027Lua" then
    error("Not a valid Lua chunk file!")
  end

  local loader = {
    d = str:sub(13), -- ヘッダはスキップ。x86 で標準的な形式を暗黙に仮定
    bytes =
      function(self, n)
        local dd = self.d
        self.d = self.d:sub(n+1)
        return dd:byte(1,n)
      end,
    -----------------------------
    byte = -- 1バイト整数
      function(self)
        return self:bytes(1)
      end,
    int  = -- 4バイト整数（リトルエンディアン）
      function(self)
        local a,b,c,d = self:bytes(4)
        return a+256*(b+256*(c+256*d))
      end,
    inst = -- 命令 (4バイト整数(ここでデコードする))
      function(self)
        return inst_decode(self:int())
      end,
    string = -- 文字列 = 長さ(size_t) ++ char * 長さ ++ \0
      function(self)
        local n = self:int()
        if n == 0 then
          return ""
        else
          local dd = self.d
          self.d = self.d:sub(n+1)
          return dd:sub(1,n-1)
        end
      end,
    number = -- 数値 = 倍精度(1+11+52)浮動小数点数
      function(self)
        local l,h,s = self:int(), self:int(), -1
        if h < 2^31 then s = 1 end
        local e = 1 + math.floor(h/2^20)%2^11 - 2^10
        if e == -1023 then
          return 0
        else
          return s*(2^e + math.ldexp(h%2^20,e-20) + math.ldexp(l,e-52))
        end
      end,
    code = -- コード = 命令の列
      function(self)
        local n = self:int()
        local code = {}
        for i=1,n do
          code[i] = self:inst()
        end
        return code
      end,
    proto = -- 関数プロトタイプ
      function(self)
        local obj = {type = type_function}
        self:string() -- 定義されたソースファイル名
        self:int()    -- 開始行
        self:int()    -- 終了行
        obj.numup  = self:byte()
        obj.numpr  = self:byte()
        obj.varflg = self:byte()
        obj.maxstk = self:byte()
        obj.code   = self:code()
        obj.kv     = self:constsV()
        obj.kp     = self:constsP()
        self:dbginfo()
        return obj
      end,
    constsV = -- 定数テーブル
      function(self)
        local n  = self:int()
        local kv = {}
        for i=1, n do
          local t = self:byte()
          if t == 0 then -- TNIL
            kv[i] = nil
          elseif t == 1 then -- TBOOLEAN
            kv[i] = self:byte()~=0
          elseif t == 3 then -- TNUMBER
            kv[i] = self:number()
          elseif t == 4 then -- TSTRING
            kv[i] = self:string()
          else
            error("Bad Tag")
          end
        end
        return kv
      end,
    constsP = -- 関数プロトタイプテーブル
      function(self)
        local n  = self:int()
        local kf = {}
        for i=1, n do
          kf[i] = self:proto()
        end
        return kf
      end,
    dbginfo = -- デバッグ情報。とりあえずスキップ
      function(self)
        local n = self:int()
        for i=1, n do
          self:int()
        end
        n = self:int()
        for i=1, n do
          self:string()
          self:int()
          self:int()
        end
        n = self:int()
        for i=1, n do
          self:string()
        end
      end,
  }
  return loader:proto()
end

-------------------------------------------------------------------------
-- Util :: オブジェクト複製
-------------------------------------------------------------------------

function dup(obj)
  local new = {}
  for k,v in pairs(obj) do new[k]=v end
  return new
end

-------------------------------------------------------------------------
-- Lib :: ライブラリ関数パッケージを返す
-------------------------------------------------------------------------

function library()
  local lib = dup(_G) -- とりあえずホスト環境のライブラリをコピー
  lib._G       = lib
  lib._VERSION = "LuLu 0.4"
  lib.arg      = {[-1]=arg[-1].." "..arg[0], [0]=arg[1], unpack(arg,2)}
  lib.type     = function(v)
                   local s = type(v)
                   if s == "table" then
                      return v.type==type_function and "function" or
                                (v.type==type_coroutine and "thread" or s)
                   else
                      return s
                   end
                 end
  lib.coroutine = { create = new_coroutine,
                    resume = co_resume,
                    yield  = co_yield,
                    wrap   = co_wrap,
                    running= co_running,
                    status = co_status }
  local function getfunc(fn)
    if type(fn)=="table" and fn.type==type_function then
      return fn
    elseif type(fn)=="table" and fn.type==type_coroutine then
      return fn.main
    else
      local rt = running_thread
      fn = fn or 1
      if fn == 0 then
        return rt.main
      else
        return rt.cstk[rt.ctop-fn].func
      end
    end
  end
  lib.getfenv = function(fn)
                  return getfunc(fn).env
                end
  lib.setfenv = function(fn, t)
                  local fn = getfunc(fn)
                  fn.env = t
                  return fn
                end
  -- 非対応
  lib.debug   = nil
  lib.module  = nil
  lib.require = nil
  lib.package = nil
  -- 現時点では非対応
  -- todo: dofile/load/loadfile/loadstring (なんとかなる気もする / 現在はホストLuaで実行される)
  -- todo: string.dump (なんとかなる気もする)
  -- todo: tostring    (of closure/thread) (めんどいだけ)
  -- get/setmetatable  (closure.__call さえ用意すればごまかせる？)
  -- pcall/xpcall      (closure.__call さえ用意すればごまかせる？)
  return lib
end

function new_closure(proto, env)
  local fn = dup(proto)
  fn.upval = {}
  fn.env   = env
  return fn
end

-------------------------------------------------------------------------
-- Coroutine :: コルーチンライブラリ
-------------------------------------------------------------------------

main_thread    = nil
running_thread = nil

function new_coroutine(fn)
  return { type = type_coroutine, status = "notstarted", main = fn }
end

STARTUP_CODE = {type=type_function, code={inst_decode(29)}} -- TAILCALL

function co_resume(co, ...)
  if co.status == "notstarted" then
    co.dstk = {nil, co.main, ...}
    co.dtop = 1 + #co.dstk
    co.cstk = {{func=STARTUP_CODE, pc=1, base=2, retpos=1, wanted=-1, uvcache={}}}
    co.ctop = 2
  elseif co.status == "suspended" then
    local a = {...}
    local narg = (co.yield_wanted==-1 and #a or co.yield_wanted)
    for i=1, narg do
      co.dstk[co.yield_from+i-1] = a[i]
    end
    co.dtop = co.yield_from + narg
  else
    error( "cannot resume ".. co.status .." coroutine" )
  end

  local rt = running_thread
  if rt then rt.status = "normal" end
  running_thread = co
  co.status = "running"
  local ret = co_run(co)
  running_thread = rt
  if rt then rt.status = "running" end

  return true, unpack(ret)
end

function co_yield(...)
  running_thread.yield_from    = YF -- てぬき
  running_thread.yield_wanted  = YW -- てぬき
  running_thread.return_values = {...}
  running_thread.status        = "suspended"
end

function co_wrap(fn) -- coroutine.wrap
  local co = new_coroutine(fn)
  return function(...)
     local a = {co_resume(co, ...)}
     return unpack(a, 2, #a)
  end
end

function co_status(co) -- coroutine.status
  return (co.status=="notstarted" and "suspended" or co.status)
end

function co_running() -- coroutine.running
  return running_thread~=main_thread and running_thread or nil
end

-------------------------------------------------------------------------
-- VM :: 関数オブジェクトを渡すとそれをメインスクリプトとして実行
-------------------------------------------------------------------------

function vm_run(proto)
  main_thread = new_coroutine(new_closure(proto,library()))
  co_resume(main_thread)
end

function co_run(co) -- == luaV_execute
  -- コールスタック
  local ci = co.cstk[co.ctop-1]
  local func,base,retpos,wanted,uvcache = ci.func, ci.base, ci.retpos, ci.wanted, ci.uvcache
  local function cipush(tail, a,b,c,d,e,f)
     ci = {func=a, pc=b, base=c, retpos=d, wanted=e, uvcache=f}
     if not tail then
       co.ctop=co.ctop+1
     end
     co.cstk[co.ctop-1]=ci
     func,base,retpos,wanted,uvcache = ci.func, ci.base, ci.retpos, ci.wanted, ci.uvcache
  end
  local function cipop()
     co.ctop=co.ctop-1
     ci = co.cstk[co.ctop-1]
     func,base,retpos,wanted,uvcache = ci.func, ci.base, ci.retpos, ci.wanted, ci.uvcache
  end

  -- データスタック
  local stk = co.dstk
  local function getreg(n)   return stk[base+n] end
  local function setreg(n,v) stk[base+n] = v    end
  local function getregs(i,n) return unpack(stk, base+i, base+i+n-1) end
  local function setregs(i,n,vals)
    local m = n
    if m > #vals then m = #vals end
    for j=1,m   do stk[base+i+j-1] = vals[j] end
    for j=m+1,n do stk[base+i+j-1] = nil  end
  end

  -- upvalueをスタックから逃がす処理
  local function closeupvals(from) 
    from = from or 0
    for k,u in pairs(uvcache) do
      if from<=k and not rawequal(u.base, u) then
        u[1]   = u.base[u.idx]
        u.base = u
        u.idx  = 1
        uvcache[k] = nil --逃がし終わったらテーブルから消す
      end
    end
  end

  local function do_call(fn, A, narg, nret, tail) -- 関数呼び出しの共通処理
     if type(fn)=="table" and fn.type==type_function then
        cipush(tail, fn, 1, base+A+1, base+A, nret, {})

        -- 実引数が仮引数より少ない場合、残りを nil で埋める
        if narg < fn.numpr then
           for i=narg, fn.numpr-1 do setreg(i, nil) end
           narg = fn.numpr
        end
        -- 可変長引数な時、必須引数を前に持ってきてbaseをずらす
        if func.varflg>0 then
           for i=0,fn.numpr-1 do
             stk[base+narg+i] = stk[base+i]
             stk[base+i]      = nil
           end
           base = base+narg
           co.cstk[co.ctop-1].base = base
        end
        -- ローカル変数領域をnilクリア
        for i=fn.numpr, fn.maxstk-1 do setreg(i,nil) end
     else
        YF, YW = A, nret -- てぬき：fnがco_yieldだった場合に使う
        local a = {fn(getregs(A+1, narg))}
        setregs(A, (nret==-1 and #a or nret), a)
        co.dtop = base+A+#a
     end
  end

  -- メインの実行ループ
  co.return_values = nil
  repeat
    local inst = func.code[ci.pc]
    ci.pc = ci.pc+1

    local OP  = inst.OP
    local A   = inst.A
    local C   = inst.C
    local B   = inst.B
    local Bx  = inst.Bx
    local sBx = inst.sBx
    local RKB if B>=2^8 then RKB = func.kv[B-2^8+1] else RKB = getreg(B) end
    local RKC if C>=2^8 then RKC = func.kv[C-2^8+1] else RKC = getreg(C) end

    if OP ==  0 then -- MOVE
       setreg(A, getreg(B))
    elseif OP == 1 then -- LOADK
       setreg(A, func.kv[Bx+1])
    elseif OP == 2 then -- LOADBOOL
       setreg(A, B~=0)
       if C~=0 then ci.pc = ci.pc+1 end
    elseif OP == 3 then -- LOADNIL
       for i=A,B do setreg(i, nil) end
    elseif OP == 4 then -- GETUPVAL
       local u = func.upval[B+1]
       setreg(A, u.base[u.idx])
    elseif OP == 5 then -- GETGLOBAL
       setreg(A, func.env[func.kv[Bx+1]])
    elseif OP == 6 then -- GETTABLE
       local rb = getreg(B)
       setreg(A, rb[RKC])
    elseif OP == 7 then -- SETGLOBAL
       func.env[func.kv[Bx+1]] = getreg(A)
    elseif OP == 8 then -- SETUPVAL
       local u = func.upval[B+1]
       u.base[u.idx] = getreg(A)
    elseif OP == 9 then -- SETTABLE
       local ra = getreg(A)
       ra[RKB] = RKC
    elseif OP == 10 then -- NEWTABLE
       setreg(A, {})
    elseif OP == 11 then -- SELF
       local rb = getreg(B)
       setreg(A,   rb[RKC])
       setreg(A+1, rb)
    elseif OP == 12 then -- ADD
       setreg(A, RKB + RKC)
    elseif OP == 13 then -- SUB
       setreg(A, RKB - RKC)
    elseif OP == 14 then -- MUL
       setreg(A, RKB * RKC)
    elseif OP == 15 then -- DIV
       setreg(A, RKB / RKC)
    elseif OP == 16 then -- MOD
       setreg(A, RKB % RKC)
    elseif OP == 17 then -- POW
       setreg(A, RKB ^ RKC)
    elseif OP == 18 then -- UNM
       setreg(A, - RKB)
    elseif OP == 19 then -- NOT
       setreg(A, not RKB)
    elseif OP == 20 then -- LEN
       setreg(A, # RKB)
    elseif OP == 21 then -- CONCAT
       local s = ""
       for i=B,C do s = s .. getreg(i) end
       setreg(A, s)
    elseif OP == 22 then -- JMP
       ci.pc = ci.pc + sBx
    elseif OP == 23 then -- EQ
       if (RKB==RKC) ~= (A~=0) then ci.pc=ci.pc+1 end
    elseif OP == 24 then -- LT
       if (RKB< RKC) ~= (A~=0) then ci.pc=ci.pc+1 end
    elseif OP == 25 then -- LE
       if (RKB<=RKC) ~= (A~=0) then ci.pc=ci.pc+1 end
    elseif OP == 26 then -- TEST
       if (not getreg(A)) == (C~=0) then ci.pc=ci.pc+1 end
    elseif OP == 27 then -- TESTSET
       if (not getreg(B)) == (C~=0) then ci.pc=ci.pc+1 else setreg(A,getreg(B)) end
    elseif OP == 28 then -- CALL : R[A],...,R[A+C-2], R[A]( R[A+1], ..., R[A+B-1] )
       do_call(getreg(A), A, B==0 and co.dtop-(base+A+1) or B-1, C-1)
    elseif OP == 29 then -- TAILCALL : return R[A]( R[A+1], ..., R[A+B-1] )
       closeupvals()
       local fn = getreg(A)
       if type(fn)=="table" and fn.type==type_function then
          local narg = B==0 and co.dtop-(base+A+1) or B-1
          setregs(retpos-base, narg+1, {getregs(A, narg+1)})
          do_call(fn, retpos-base, narg, C-1, true)
       else
          local a = {fn(getregs(A+1, B==0 and co.dtop-(base+A+1) or B-1))}
          if co.ctop == 2 then
            co.status = "dead"
            co.return_values = a
          else
            setregs(retpos-base, (wanted==-1 and #a or wanted), a)
            co.dtop = retpos + #a
            cipop()
          end
       end
    elseif OP == 30 then -- RETURN : R[A], ..., R[A+B-1]
       closeupvals()
       local numret = B==0 and co.dtop-(base+A) or B-1
       if co.ctop == 2 then
         co.status = "dead"
         co.return_values = {getregs(A, numret)}
       else
         setregs(retpos-base, (wanted==-1 and numret or wanted), {getregs(A,numret)})
         co.dtop = retpos + numret
         cipop()
       end
    elseif OP == 31 then -- FORLOOP : for i=a,b,c 型ループ
       local idx = getreg(A)+getreg(A+2)
       if (0<getreg(A+2)) == (idx<=getreg(A+1)) then
         ci.pc = ci.pc + sBx
         setreg(A,   idx) -- ループ用の内部カウンタ
         setreg(A+3, idx) -- Luaのコードから見える変数
       end
    elseif OP == 32 then -- FORPREP : ループ開始
       setreg(A, getreg(A)-getreg(A+2))
       ci.pc = ci.pc + sBx
    elseif OP == 33 then -- TFORLOOP : for _ in R[A] 型ループ
       func.code[ci.pc].OP = 999 -- 無理矢理実装／動的命令書き換え JMP(22)-> 999
       setreg(A+3, getreg(A))
       setreg(A+4, getreg(A+1))
       setreg(A+5, getreg(A+2))
       do_call(getreg(A+3), A+3, 2, C)
      elseif OP == 999 then -- TFORLOOPの直後のJMP
       A = func.code[ci.pc-2].A+3
       if getreg(A) ~= nil then
         setreg(A-1, getreg(A))
         ci.pc = ci.pc + sBx
       end
    elseif OP == 34 then -- SETLIST : RA[(C-1)*FPF+i] = R[A+i] for 1<=i<=B
       local FPF = 50 -- 本家Luaのソースではマクロで定義されてる
       if B==0 then B = co.dtop-(base+A)-1 end
       if C==0 then C = func.code[ci.pc].rawval ci.pc=ci.pc+1 end
       local arr = getreg(A)
       for i=1,B do
          arr[(C-1)*FPF+i] = getreg(A+i)
       end
    elseif OP == 35 then -- CLOSE
       closeupvals(A)
    elseif OP == 36 then -- CLOSURE
       local cl = new_closure(func.kp[Bx+1], func.env)
       for i=1, cl.numup do
         if func.code[ci.pc].OP == 0 then -- MOVE : 現在のスタックの変数をupval化
            local b = func.code[ci.pc].B
            local uv = uvcache[b]
            if not uv then uv={base=stk, idx=base+b} uvcache[b]=uv end
            cl.upval[i] = uv
         else -- GETUPVAL : 親(祖先)のスタックのupvalを持ってくる
            cl.upval[i] = func.upval[xB()+1]
         end
         ci.pc = ci.pc+1
       end
       setreg(A, cl)
    elseif OP == 37 then -- VARARG : R[A],...,R[A+B-2] = ...
       local numvar = base-retpos-1-func.numpr
       if B == 0 then -- multivalue
          B = numvar
          co.dtop = base + A + numvar
       else
          B = B-1
       end
       for i=0,B-1 do
         if i < numvar then
           setreg(A+i, stk[base-numvar+i])
         else
           setreg(A+i, nil)
         end
       end
    else
       error("Unknown VM Instruction")
    end
  until co.return_values
  return co.return_values
end

-------------------------------------------------------------------------
-- Main
-------------------------------------------------------------------------

if arg[1] then
   local chunk = string.dump(loadfile(arg[1]))
   -- local fh    = io.open(arg[1],"rb")
   -- local chunk = fh:read("*a")
   -- fh:close()
   vm_run( load_chunk(chunk) )
else
   print( "Usage: lua lulu.lua [script.lua] [args]" )
end
